import array
import sys
import usb.core
import usb.util
import multiprocessing as mp
import time
import hashlib
import enum
import atexit
from typing import List, Optional
from .packet import Packet, PacketDestination
from .common_payload import RequestDeviceIdPayload, ReplyDeviceIdPayload, RequestDeviceNamePayload, ReplyDeviceNamePayload
from .node import Node
from .connection import Connection

USB_HID_REPORT_LENGTH = 64


class USBConnection(Connection):
    """An implementation of :class:`amfiprot.Connection` used to connect to USB HID devices."""
    MAX_PAYLOAD_SIZE = 54  # 1 byte needed for CRC

    def __init__(self, vendor_id: int, product_id: int, serial_number: str = None):
        """ If no serial number is given, the first device that matches vendor_id and product_id is used. """
        self.vendor_id = vendor_id
        self.product_id = product_id
        self.usb_serial_number = serial_number

        self.usb_device = get_matching_device(vendor_id, product_id, serial_number)

        if self.usb_device is None:
            raise ConnectionError("Could not connect to device!")

        self.receive_process: mp.Process = None
        self.transmit_process: mp.Process = None
        self.usb_task: mp.Process = None
        self.nodes: List[Node] = []
        self.transmit_queue: mp.Queue = mp.Queue()
        self.global_receive_queue: mp.Queue = mp.Queue()
        self.usb_connection_lost: mp.Event = mp.Event()
        self.node_update_queue: mp.Queue = mp.Queue()

    def __del__(self):
        try:
            self.stop()
        except AttributeError:
            pass  # Processes not started

    @classmethod
    def discover(cls):
        devices = usb.core.find(find_all=True)
        device_list = []

        for device in devices:
            try:
                if device.manufacturer is None or device.product is None:
                    raise ValueError("Unknown device")

                try:
                    serial_number = device.serial_number
                except Exception:
                    serial_number = None

                device_info = {'vid': device.idVendor, 'pid': device.idProduct, 'manufacturer': device.manufacturer, 'product': device.product, 'serial_number': serial_number}
                device_list.append(device_info)
            except Exception:
                continue

        return device_list

    def find_nodes(self) -> List[Node]:
        # TODO: Clean this method. Implement helper functions for blocking send/receive

        # Create 'request device id' packet
        payload = RequestDeviceIdPayload()
        packet = Packet.from_payload(payload, destination_id=PacketDestination.BROADCAST)

        # Send packet via USB
        bytes_to_transmit: array.array = array.array('B', [1])  # USB header Report ID, packet length only required on IN packets (???)
        bytes_to_transmit.extend(packet.to_bytes())

        while len(bytes_to_transmit) < 64:
            bytes_to_transmit.append(0)

        self.usb_device.write(0x1, bytes_to_transmit, 1000)

        start_time = time.time()

        rx_packets = []
        while time.time() - start_time < 1:
            data = None

            try:
                data = self.usb_device.read(0x81, 64)
            except usb.core.USBTimeoutError:
                #print("USB timed out while waiting for Device ID reply packets.")
                continue

            if len(data) == 0:  # Extra safe guard for Linux (or previous libusb version??)
                continue

            rx_packet = Packet(data[2:])
            #print(rx_packet)

            if type(rx_packet.payload) == ReplyDeviceIdPayload:
                rx_packets.append(rx_packet)

        nodes = []

        for packet in rx_packets:
            if type(packet.payload) == ReplyDeviceIdPayload:
                uuids = [node.uuid for node in nodes]
                if packet.payload.uuid not in uuids:
                    node = Node(tx_id=packet.payload.tx_id, uuid=packet.payload.uuid, connection=self)
                    nodes.append(node)

        # Request device name for all found nodes
        for node in nodes:
            payload = RequestDeviceNamePayload()
            packet = Packet.from_payload(payload, destination_id=node.tx_id)

            # Send packet via USB
            bytes_to_transmit: array.array = array.array('B', [1])  # USB header Report ID, packet length only required on IN packets (???)
            bytes_to_transmit.extend(packet.to_bytes())

            while len(bytes_to_transmit) < 64:
                bytes_to_transmit.append(0)

            self.usb_device.write(0x1, bytes_to_transmit, 1000)

            start_time = time.time()

            while time.time() - start_time < 1:
                data = None

                try:
                    data = self.usb_device.read(0x81, 64)
                except usb.core.USBTimeoutError:
                    # print("USB timed out while waiting for Device ID reply packets.")
                    continue

                rx_packet = Packet(data[2:])
                # print(rx_packet)

                if type(rx_packet.payload) == ReplyDeviceNamePayload:
                    node.name = rx_packet.payload.name
                    break


        # Restart connection if nodes list changed
        if nodes_changed(nodes, self.nodes):
            self.nodes = nodes
            # Send updated node list to RX/TX processes (if started)
            if self.receive_process is not None and self.transmit_process is not None:
                # HACK: just restart connection using new self.nodes list
                self.stop()
                self.start()
            # print("New node detected!")

        return self.nodes

    def enqueue_packet(self, packet: Packet):
        self.transmit_queue.put(packet)

    def max_payload_size(self) -> int:
        return self.MAX_PAYLOAD_SIZE

    def start(self):
        usb_device_hash = generate_device_hash(self.usb_device)
        self.usb_device.reset()
        usb.util.dispose_resources(self.usb_device)
        del self.usb_device

        atexit.register(connection_exit_handler, self)

        # Create rx process
        tx_ids = [node.tx_id for node in self.nodes]
        rx_queues = [node.receive_queue for node in self.nodes]

        self.usb_task = mp.Process(target=usb_task, args=(usb_device_hash, tx_ids, rx_queues, self.transmit_queue, self.global_receive_queue, self.node_update_queue))
        self.usb_task.start()
        time.sleep(1)  # To allow processes to start up

    def stop(self):
        # TODO: Send stop request to task and wait for acknowledge (allows outbound packets to be sent before stopping)

        if self.usb_task is not None:
            if self.usb_task.is_alive():
                time.sleep(1)  # To allow pending tx packets to be sent
                self.usb_task.terminate()
                self.usb_task.join()

    def refresh(self):
        #rx_queues = [node.receive_queue for node in self.nodes]  # Cannot send queues in a queue to other process
        tx_ids = [node.tx_id for node in self.nodes]
        self.node_update_queue.put({'tx_ids': tx_ids})

    def __str__(self):
        bus = self.usb_device.bus
        address = self.usb_device.address

        vendor_id = self.usb_device.idVendor
        product_id = self.usb_device.idProduct

        manufacturer_string_index = self.usb_device.iManufacturer
        product_string_index = self.usb_device.iProduct
        serial_number_string_index = self.usb_device.iSerialNumber

        manufacturer = usb.util.get_string(self.usb_device, manufacturer_string_index)
        product = usb.util.get_string(self.usb_device, product_string_index)
        serial_number = usb.util.get_string(self.usb_device, serial_number_string_index)

        return f"{product} ({manufacturer}) on bus {bus} address {address}. VID=0x{vendor_id:04X},"\
               f" PID=0x{product_id:04X}, SN={serial_number}"


class ConnectionState(enum.IntEnum):
    INIT = 0
    CONNECTED = 1
    DISCONNECTED = 2


def usb_task(usb_device_hash, tx_ids, rx_queues: List[mp.Queue], tx_queue: mp.Queue, global_receive_queue: mp.Queue, node_update_queue: mp.Queue):
    IN_ENDPOINT = 0x81
    OUT_ENDPOINT = 0x01
    RETRY_LIMIT = 10

    retry_count = 0
    dev = None

    while dev is None:
        dev = get_usb_device_by_hash(usb_device_hash)
        state = ConnectionState.CONNECTED

        retry_count = retry_count + 1
        if retry_count > RETRY_LIMIT and dev is None:
            raise ConnectionError("Subprocess could not find device.")

    tx_ids_local = tx_ids
    rx_queues_local = rx_queues

    while True:
        if state == ConnectionState.CONNECTED:

            # Send all pending packets
            while not tx_queue.empty():
                tx_packet = tx_queue.get_nowait()

                byte_data = array.array('B', [1])
                byte_data.extend(tx_packet.to_bytes())
                byte_data.extend([0] * (64 - len(byte_data)))
                try:
                    bytes_written = dev.write(OUT_ENDPOINT, byte_data, timeout=1000) # TODO: Do something about this timeout! It was 1 ms before in order not to block reading, but sometimes it is not enough time to actually send the packet.
                except usb.core.USBError as e:  # TODO: Check disconnect in some other way before getting from tx_queue, because this drops packets!
                    print(f"Could not send packet ({e})")
                    continue

            # Check for tx_id change before receiving
            if not node_update_queue.empty():
                update = node_update_queue.get()
                tx_ids_local = update['tx_ids']

            # Try to receive
            try:
                rx_data = dev.read(IN_ENDPOINT, USB_HID_REPORT_LENGTH, timeout=0)
            except usb.core.USBTimeoutError as e:
                # print(e)
                continue
            except usb.core.USBError as e:
                print("USB connection lost.")
                usb.util.dispose_resources(dev)
                del dev
                state = ConnectionState.DISCONNECTED
                continue

            if len(rx_data) == 0:
                continue

            rx_packet = Packet(rx_data[2:])

            if global_receive_queue.full():
                print("Global receive queue full! Packet discarded.")
            else:
                global_receive_queue.put_nowait(
                    rx_packet)  # TODO: What if queue is full? Should probably only keep newest packets

            # Push packet to correct rx_queue
            try:
                index = tx_ids_local.index(rx_packet.source_id)  # .index returns ValueError if value not found in list
                if rx_queues_local[index].full():
                    print(f"RX queue [TxID {tx_ids_local[index]}] full! Packet discarded.")
                else:
                    rx_queues_local[index].put_nowait(rx_packet)

            except ValueError:
                # print(f"Packet TxID {rx_packet.source_id} does not match any nodes.")
                pass

        elif state == ConnectionState.DISCONNECTED:
            print("Reconnecting...")

            dev = get_usb_device_by_hash(usb_device_hash)

            if dev is not None:
                print("Connection re-established!")
                state = ConnectionState.CONNECTED
            else:
                time.sleep(1)
        else:
            raise ValueError("Invalid state in USB task.")


def connect_usb(vendor_id, product_id, serial_number=None):
    for i in range(3):
        try:
            devices = get_usb_devices(vendor_id, product_id)

            if len(devices) > 0:
                break
            else:
                time.sleep(1)
                continue

        except ValueError:
            print("Device not found. Retrying in 1 sec...")
            time.sleep(1)

    if len(devices) < 1:
        raise ConnectionError("Could not reconnect")

    dev = devices[0]

    if "linux" in sys.platform.lower():
        for config in dev:
            for i in range(config.bNumInterfaces):
                if dev.is_kernel_driver_active(i):
                    dev.detach_kernel_driver(i)

    return dev


def get_usb_devices(vendor_id, product_id):
    return list(usb.core.find(find_all=True, idVendor=vendor_id, idProduct=product_id))


def get_serial_number(device: usb.core.Device) -> int:
    return usb.util.get_string(device, device.iSerialNumber)


def get_matching_device(vendor_id, product_id, serial_number) -> usb.core.Device:
    if serial_number is None:
        device: usb.core.Device = usb.core.find(idVendor=vendor_id, idProduct=product_id)

        if device is None:
            raise ConnectionError("No match!")

        device.reset()

        if "linux" in sys.platform.lower():
            # print("Host is Linux")
            linux_usb_workaround(device)
        # device.set_configuration()
        return device
    else:
        devices = usb.core.find(find_all=True, idVendor=vendor_id, idProduct=product_id)

        for device in devices:
            if get_serial_number(device) == serial_number:
                return device
        return None


def generate_device_hash(device: usb.core.Device) -> str:
    string_to_hash = str(device.idVendor) + str(device.idProduct) + str(device.manufacturer) + str(device.product) + str(device.serial_number)
    encoded_string = string_to_hash.encode('utf-8')

    hash = hashlib.md5()
    hash.update(encoded_string)
    return hash.hexdigest()


def get_usb_device_by_hash(hash: str) -> Optional[usb.core.Device]:
    devices = list(usb.core.find(find_all=True))

    if len(devices) > 0:
        for device in devices:
            try:
                current_hash = generate_device_hash(device)
            except ValueError:
                continue
            except NotImplementedError:
                continue

            if hash == current_hash:

                # if "linux" in sys.platform.lower():
                #     print("Host is Linux!")
                #     for config in device:
                #         for i in range(config.bNumInterfaces):
                #             if device.is_kernel_driver_active(i):
                #                 device.detach_kernel_driver(i)
                # device.set_configuration()

                return device

    return None


def nodes_changed(list1: List[Node], list2):
    # Check if lengths differ
    if len(list1) != len(list2):
        return True

    # Sort both lists by uuid and compare each element (list equality will not work)
    list1.sort(key=lambda x: x.uuid)
    list2.sort(key=lambda x: x.uuid)

    for (item1, item2) in zip(list1, list2):
        if item1.uuid != item2.uuid:
            return True

    return False


def linux_usb_workaround(device):
    for config in device:
        for i in range(config.bNumInterfaces):
            if device.is_kernel_driver_active(i):
                device.detach_kernel_driver(i)


def connection_exit_handler(connection):
    connection.stop()



import array
import sys
import serial
import serial.tools.list_ports
import multiprocessing as mp
import time
import hashlib
import enum
import atexit
from typing import List, Optional
from cobs import cobs
from .packet import Packet, PacketDestination
from .common_payload import RequestDeviceIdPayload, ReplyDeviceIdPayload, RequestDeviceNamePayload, ReplyDeviceNamePayload
from .node import Node
from .connection import Connection

class UARTConnection(Connection):
    """An implementation of :class:`amfiprot.Connection` used to connect to UART devices."""
    MAX_PAYLOAD_SIZE = 54  # 1 byte needed for CRC

    def __init__(self, port: str, baudrate: int = 115200):
        self.port = port
        self.baudrate = baudrate
        self.serial_device = get_matching_device(port, baudrate)

        if self.serial_device is None:
            raise ConnectionError("Could not connect to device!")

        self.receive_process: mp.Process = None
        self.transmit_process: mp.Process = None
        self.uart_task: mp.Process = None
        self.nodes: List[Node] = []
        self.transmit_queue: mp.Queue = mp.Queue()
        self.global_receive_queue: mp.Queue = mp.Queue()
        self.uart_connection_lost: mp.Event = mp.Event()
        self.node_update_queue: mp.Queue = mp.Queue()
        

    def __del__(self):
        try:
            self.stop()
        except AttributeError:
            pass  # Processes not started

    @classmethod
    def discover(cls):
        ports = serial.tools.list_ports.comports()
        device_list = []

        for port in ports:
            device_info = {
                'port': port.device,
                'description': port.description,
                'hwid': port.hwid
            }
            device_list.append(device_info)

        return device_list

    def find_nodes(self) -> List[Node]:
        payload = RequestDeviceIdPayload()
        packet = Packet.from_payload(payload, destination_id=PacketDestination.BROADCAST)

        bytes_to_transmit: array.array = array.array('B')
        bytes_to_transmit.extend(packet.to_bytes())
        cobs_encoded = cobs.encode(bytes(bytes_to_transmit)) + b'\x00'

        self.serial_device.write(cobs_encoded)

        start_time = time.time()

        rx_packets = []
        while time.time() - start_time < 1:
            if self.serial_device.in_waiting > 0:
                data = self.serial_device.read_until(b'\x00')
                try:
                    cobs_decoded = cobs.decode(data[:-1])
                    arr = array.array('B') 
                    arr.frombytes(cobs_decoded)
                    rx_packet = Packet(arr)
                    if type(rx_packet.payload) == ReplyDeviceIdPayload:
                        rx_packets.append(rx_packet)
                except:
                    pass
                    #print("UART package error")

        nodes = []

        for packet in rx_packets:
            if type(packet.payload) == ReplyDeviceIdPayload:
                uuids = [node.uuid for node in nodes]
                if packet.payload.uuid not in uuids:
                    node = Node(tx_id=packet.payload.tx_id, uuid=packet.payload.uuid, connection=self)
                    nodes.append(node)

        for node in nodes:
            payload = RequestDeviceNamePayload()
            packet = Packet.from_payload(payload, destination_id=node.tx_id)

            bytes_to_transmit: array.array = array.array('B')
            bytes_to_transmit.extend(packet.to_bytes())
            cobs_encoded = cobs.encode(bytes(bytes_to_transmit)) + b'\x00'

            self.serial_device.write(cobs_encoded)

            start_time = time.time()

            while time.time() - start_time < 1:
                if self.serial_device.in_waiting > 0:
                    data = self.serial_device.read_until(b'\x00')
                    try:
                        cobs_decoded = cobs.decode(data[:-1])
                        arr = array.array('B') 
                        arr.frombytes(cobs_decoded)
                        rx_packet = Packet(arr)
                        if type(rx_packet.payload) == ReplyDeviceNamePayload:
                            node.name = rx_packet.payload.name
                            break
                    except:
                        pass
                        #print("UART package error")

        if nodes_changed(nodes, self.nodes):
            self.nodes = nodes
            if self.receive_process is not None and self.transmit_process is not None:
                self.stop()
                self.start()

        self.serial_device.close()

        return self.nodes

    def enqueue_packet(self, packet: Packet):
        self.transmit_queue.put(packet)

    def max_payload_size(self) -> int:
        return self.MAX_PAYLOAD_SIZE

    def start(self):
        atexit.register(connection_exit_handler, self)

        tx_ids = [node.tx_id for node in self.nodes]
        rx_queues = [node.receive_queue for node in self.nodes]

        self.uart_task = mp.Process(target=uart_task, args=(self.port, self.baudrate, tx_ids, rx_queues, self.transmit_queue, self.global_receive_queue, self.node_update_queue))
        self.uart_task.start()
        time.sleep(1)  # To allow processes to start up

    def stop(self):
        if self.uart_task is not None:
            if self.uart_task.is_alive():
                time.sleep(1)
                self.uart_task.terminate()
                self.uart_task.join()
                self.serial_device.close()

    def refresh(self):
        tx_ids = [node.tx_id for node in self.nodes]
        self.node_update_queue.put({'tx_ids': tx_ids})

    def __str__(self):
        return f"UART Connection on port {self.port} at baudrate {self.baudrate}"


class ConnectionState(enum.IntEnum):
    INIT = 0
    CONNECTED = 1
    DISCONNECTED = 2


def uart_task(port, baudrate, tx_ids, rx_queues: List[mp.Queue], tx_queue: mp.Queue, global_receive_queue: mp.Queue, node_update_queue: mp.Queue):
    RETRY_LIMIT = 10
    retry_count = 0
    dev = None

    while dev is None:
        try:
            dev = serial.Serial(port, baudrate, timeout=1)
            state = ConnectionState.CONNECTED
        except serial.SerialException:
            retry_count += 1
            if retry_count > RETRY_LIMIT:
                raise ConnectionError("Subprocess could not find device.")
            time.sleep(1)

    tx_ids_local = tx_ids
    rx_queues_local = rx_queues

    while True:
        if state == ConnectionState.CONNECTED:
            while not tx_queue.empty():
                tx_packet = tx_queue.get_nowait()
                byte_data = array.array('B')
                byte_data.extend(tx_packet.to_bytes())
                cobs_encoded = cobs.encode(bytes(byte_data)) + b'\x00'
                try:
                    dev.write(cobs_encoded)
                except serial.SerialException as e:
                    print(f"Could not send packet ({e})")
                    continue

            if not node_update_queue.empty():
                update = node_update_queue.get()
                tx_ids_local = update['tx_ids']

            try:
                if dev.in_waiting > 0:
                    data = dev.read_until(b'\x00')
                    try:
                        cobs_decoded = cobs.decode(data[:-1])
                        arr = array.array('B') 
                        arr.frombytes(cobs_decoded)
                        rx_packet = Packet(arr)
                    
                        if global_receive_queue.full():
                            print("Global receive queue full! Packet discarded.")
                        else:
                            global_receive_queue.put_nowait(rx_packet)

                        index = tx_ids_local.index(rx_packet.source_id)
                        if rx_queues_local[index].full():
                            print(f"RX queue [TxID {tx_ids_local[index]}] full! Packet discarded.")
                        else:
                            rx_queues_local[index].put_nowait(rx_packet)

                    except:
                        pass
                        #print("UART package error")

            except serial.SerialException as e:
                print("UART connection lost.")
                dev.close()
                state = ConnectionState.DISCONNECTED
                continue

        elif state == ConnectionState.DISCONNECTED:
            print("Reconnecting...")
            try:
                dev = serial.Serial(port, baudrate, timeout=1)
                state = ConnectionState.CONNECTED
            except serial.SerialException:
                time.sleep(1)
        else:
            raise ValueError("Invalid state in UART task.")


def get_matching_device(port, baudrate):
    try:
        dev = serial.Serial(port, baudrate, timeout=1)
        return dev
    except serial.SerialException:
        return None


def nodes_changed(list1: List[Node], list2):
    if len(list1) != len(list2):
        return True

    list1.sort(key=lambda x: x.uuid)
    list2.sort(key=lambda x: x.uuid)

    for item1, item2 in zip(list1, list2):
        if item1.uuid != item2.uuid:
            return True

    return False


def connection_exit_handler(connection):
    connection.stop()


from __future__ import annotations
import multiprocessing as mp
import time
import typing
from .packet import Packet, PacketType
from .payload import Payload

if typing.TYPE_CHECKING:
    from .connection import Connection


class Node:
    """ A `Node` represents a single endpoint on a `Connection`. One `Connection` can
    have multiple nodes, e.g. if the PC connects via USB to a device which in turn is connected
    to additional devices via RF. """
    def __init__(self, tx_id, uuid, connection: Connection):
        self.connection = connection
        self.tx_id = tx_id
        self.uuid = uuid
        self.receive_queue: mp.Queue = mp.Queue()
        self.packet_number = 0
        self.name = None

    def packet_available(self) -> bool:
        return not self.receive_queue.empty()  # Do not use qsize() as it is unreliable, and not supported on Mac

    def get_packet(self, blocking=False, timeout_ms=1000) -> typing.Optional[Packet]:
        if self.receive_queue.empty() and not blocking:
            return None

        # TODO: Implement timeout

        return self.receive_queue.get()

    def send_packet(self, packet: Packet):
        """ Send a pre-assembled packet. Note that this does not increment the packet number! """
        self.connection.enqueue_packet(packet)

    def send_payload(self,
                     payload: Payload,
                     source_id: int = 0,
                     packet_type: PacketType = PacketType.NO_ACK):
        packet = Packet.from_payload(payload,
                                     destination_id=self.tx_id,
                                     source_id=source_id,
                                     packet_type=packet_type,
                                     packet_number=self.packet_number)
        self.send_packet(packet)

        self.packet_number = (self.packet_number + 1) % 255

    def flush_receive_queue(self):
        while not self.receive_queue.empty():
            self.receive_queue.get_nowait()

    def max_payload_size(self):
        return self.connection.max_payload_size()

    def __str__(self):
        return f"<Node> name: {self.name}, tx_id: {self.tx_id}, uuid: 0x{self.uuid:024X}"



__all__ = ['connection', 'usb_connection', 'uart_connection', 'device', 'packet', 'payload', 'common_payload']
__version__ = '0.1.8'

from .payload import Payload, PayloadType
from .node import Node
from .packet import Packet
from .device import Device
from .connection import Connection
from .usb_connection import USBConnection
from .uart_connection import UARTConnection
from .common_payload import *



import warnings
from typing import List, Union, Tuple
from .common_payload import *


class Configurator:
    def __init__(self, device):
        self.device = device

    def read_all(self, flat_list: bool = False) -> List[dict]:
        config = []

        for cat_index in range(self._get_category_count()):
            category_name = self._get_category_name(cat_index)
            category = {'name': category_name, 'parameters': []}

            for param_index in range(self._get_parameter_count(cat_index)):
                name, uid = self._get_parameter_name_uid(cat_index, param_index)
                value = self.read(uid)

                category['parameters'].append({'uid': uid, 'name': name, 'value': value})

            if flat_list:
                for param in category['parameters']:
                    param['category'] = category_name
                    config.append(param)
            else:
                config.append(category)
        return config

    def write_all(self, config):
        # Detect if config is a flat list or categories with nested parameters
        flat_list = False

        if 'uid' in config[0].keys():
            print("Flat list detected!")
            flat_list = True

        # Write all parameters
        if flat_list:
            for parameter in config:
                try:
                    self.write(parameter['uid'], parameter['value'])
                except ValueError:
                    warnings.warn(f"Parameter \"{parameter['name']}\" ({parameter['uid']}) does not exist on device")
        else:
            for category in config:
                for parameter in category['parameters']:
                    try:
                        self.write(parameter['uid'], parameter['value'])
                    except ValueError:
                        warnings.warn(f"Parameter \"{parameter['name']}\" ({parameter['uid']}) does not exist on device")

    def read(self, uid, return_datatype: bool = False) -> Union[int, float, bool, str]:
        self.device.node.send_payload(RequestConfigurationValueUidPayload(uid))
        packet = self.device._await_packet(ReplyConfigurationValueUidPayload)

        if packet.payload.uid != uid:  # Does this ever happen?
            warnings.warn("Config UIDs did not match. Trying again...")
            packet = self.device._await_packet(ReplyConfigurationValueUidPayload)

        if return_datatype:
            return packet.payload.config_value, packet.payload.data_type
        else:
            return packet.payload.config_value

    def write(self, uid, value) -> Union[int, float, bool, str]:
        try:
            old_value, data_type = self.read(uid, return_datatype=True)
        except TimeoutError:
            raise ValueError(f"Parameter does not exist on target (UID: {uid}).")

        if type(value) != type(old_value):
            raise ValueError(f"Data type mismatch (given {type(value)}, expected {type(old_value)}).")

        self.device.node.send_payload(SetConfigurationValueUidPayload(uid, value, data_type))
        response = self.device._await_packet(ReplyConfigurationValueUidPayload)

        return response.payload.config_value

    def reset_to_default(self):
        self.device.node.send_payload(LoadDefaultConfigurationPayload())

    def _get_category_count(self):
        self.device.node.send_payload(RequestCategoryCountPayload())
        packet = self.device._await_packet(ReplyCategoryCountPayload)
        return packet.payload.category_count

    def _get_category_name(self, index) -> str:
        self.device.node.send_payload(RequestConfigurationCategoryPayload(index))
        packet = self.device._await_packet(ReplyConfigurationCategory)
        return packet.payload.category_name

    def _get_parameter_count(self, index):
        self.device.node.send_payload(RequestConfigurationValueCountPayload(index))
        packet = self.device._await_packet(ReplyConfigurationValueCountPayload)
        return packet.payload.config_value_count

    def _get_parameter_name_uid(self, category_index, parameter_index) -> Tuple[str, int]:
        self.device.node.send_payload(RequestConfigurationNameUidPayload(category_index, parameter_index))
        packet = self.device._await_packet(ReplyConfigurationNameUidPayload)
        return packet.payload.configuration_name, packet.payload.configuration_uid

    def __save_current_config_as_default(self):
        raise NotImplementedError
